use num::FromPrimitive;
use std::io;
// Used num & std

enum_from_primitive!{
	pub enum Instruction {
		PUSH    = 0x00,
		POP     = 0x01,
		TAKE    = 0x02,
		MEMFREE = 0x03,
		// Stack actions

		ADD     = 0x04,
		SUBS    = 0x05,
		MULT    = 0x06,
		DIV     = 0x07,
		REMN    = 0x08,
		// Operators

		EQUAL   = 0x09,
		NEQUAL  = 0x0A,
		GREATER = 0x0B,
		LOWER   = 0x0C,
		// Comparators

		AND     = 0x0D,
		OR      = 0x0E,
		XOR     = 0x0F,
		NOT     = 0x10,
		// Logic operators

		INTOUT  = 0x11,
		CHAROUT = 0x12,
		STROUT  = 0x13,
		NEWLINE = 0x14,
		// Output actions

		CHARIN  = 0x15,
		STRIN   = 0x16,
		// Input actions

		IF      = 0x17,
		WHILE   = 0x18,
		LOOP    = 0x19,
		END     = 0x1A,
		// Code blocks

		VER     = 0x1B,
		LICENSE = 0x1C,
		// VM info

		PANIC   = 0x1D,
		THROW   = 0x1E,
		EXIT    = 0x1F
		// Exit actions
	}
}

pub fn jBrain_version() -> String {
	return "0.1.1".to_string();
}

pub fn jBrain_license() -> String {
	let a = "MIT License";
	let b = "Copyright (c) 2020 Henry Díaz";
	let c = "Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the \"Software\"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:";
	let d = "The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.";
	let e = "THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.";

	return format!("{}\n\n{}\n\n{}\n\n{}\n\n{}", a, b, c, d, e);
}

pub struct jBrainVM {
	push: bool,
	take: bool,
	block: i32,
	stack: Vec<u8>
}

impl jBrainVM {
	pub fn new() -> jBrainVM {
		println!("Starting jBrain instance...");
		return jBrainVM{ push: false, take: false, block: 0, stack: vec![] };
	}

	pub fn interpreter(&mut self, bytecode: String) -> () {
		for execbyte in bytecode.chars() {
			self.execute(execbyte as u8);
		}
	}
	fn execute(&mut self, execbyte: u8) -> () {
		if self.push {
			self.push(execbyte);
            self.push = false;

		} else if self.take {
			self.take(execbyte);
			self.take = false;

		} else {
			let op: Option<Instruction> = Instruction::from_u8(execbyte);
			match op {
				None => {
					panic!("Unknown instruction, ERROR 0x0001");
				},
				Some(bc) => {
					match bc {
						Instruction::PUSH => {
							self.push = true;
						},

						Instruction::POP => {
							self.pop();
						},

						Instruction::TAKE => {
							self.take = true;
						},

						Instruction::MEMFREE => {
							self.stack.clear();
						},
						// Stack actions

						Instruction::ADD => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a + b;
							self.push(c as u8);
						},

						Instruction::SUBS => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a - b;
							self.push(c as u8);
						},

						Instruction::MULT => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a * b;
							self.push(c as u8);
						},

						Instruction::DIV => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a / b;
							self.push(c as u8);
						},

						Instruction::REMN => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;
							let c = a % b;
							self.push(c as u8);
						},
						// Operators

						Instruction::EQUAL => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;

							let c: u8;

							if a == b {
								c = 1;
							} else {
								c = 0;
							}

							self.push(c);
						}

						Instruction::NEQUAL => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;

							let c: u8;

							if a != b {
								c = 1;
							} else {
								c = 0;
							}

							self.push(c);
						}

						Instruction::GREATER => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;

							let c: u8;

							if a > b {
								c = 1;
							} else {
								c = 0;
							}

							self.push(c);
						}

						Instruction::LOWER => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;

							let c: u8;

							if a < b {
								c = 1;
							} else {
								c = 0;
							}

							self.push(c);
						}
						// Comparators

						Instruction::AND => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;

							let c: u8;

							if a != 0 && b != 0 {
								c = 1;
							} else {
								c = 0;
							}

							self.push(c);
						}

						Instruction::OR => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;

							let c: u8;

							if a != 0 || b != 0 {
								c = 1;
							} else {
								c = 0;
							}

							self.push(c);
						}

						Instruction::XOR => {
							let a = self.pop() as i32;
							let b = self.pop() as i32;

							let c: u8;

							if (a != 0 || b != 0) && !(a != 0 && b != 0) {
								c = 1;
							} else {
								c = 0;
							}

							self.push(c);
						}

						Instruction::NOT => {
							let a = self.pop() as i32;

							self.push(!a as u8);
						}
						// Logic operators

						Instruction::INTOUT => {
							print!("{}", self.pop() as i32);
						},

						Instruction::CHAROUT => {
							print!("{}", self.pop() as char);
						},

						Instruction::STROUT => {
							let mut output = String::new();
							for i in 0..self.stack.len() {
								output = format!("{}{}", output, self.stack[i] as char);
							}
							print!("{}", output);
						},

						Instruction::NEWLINE => {
							print!("\n");
						},
						// Output actions

						Instruction::CHARIN => {
							let mut strin: String = String::new();

							io::stdin().read_line(&mut strin).unwrap();
							let input_chars: Vec<char> = strin.chars().collect();

							self.push(format!("{}", input_chars[0]).parse().unwrap());
						},

						Instruction::STRIN => {
							let mut strin: String = String::new();

							io::stdin().read_line(&mut strin).unwrap();

							for letter in strin.chars() {
								self.push(format!("{}", letter).parse().unwrap());
							}
						},
						// Input actions

						Instruction::IF => {
							self.block += 1;
							self.execute(execbyte);
							let stack_last = self.pop();

							if stack_last != 0 {
								self.stack.push(stack_last);
								
								while execbyte != (Instruction::END as u8) {
									self.execute(execbyte);
								}
							}
						}

						Instruction::WHILE => {
							self.block += 1;
							self.execute(execbyte);
							let stack_last = self.pop();

							let mut instructions: Vec<u8> = Vec::new();

							if stack_last != 0 {
								while execbyte != (Instruction::END as u8) {
									instructions.push(execbyte);
								}
							}

							while stack_last != 0 {
								for instruction in instructions.iter() {
									self.execute(*instruction);
								}
							}
						}

						Instruction::LOOP => {
							self.block += 1;
							self.execute(execbyte);
							let stack_last = self.pop();

							let mut instructions: Vec<u8> = Vec::new();

							if stack_last != 0 {
								while execbyte != (Instruction::END as u8) {
									instructions.push(execbyte);
								}
							}

							loop {
								for instruction in instructions.iter() {
									self.execute(*instruction);
								}
							}
						}

						Instruction::END => {
							self.block -= 1;
						}
						// Code blocks

						Instruction::VER => {
							println!("jBrainVM v{}", jBrain_version());
						},

						Instruction::LICENSE => {
							println!("{}", jBrain_license());
						},
						// VM info

						Instruction::PANIC => {
							panic!("This program crashed with ERROR 0x0002");
						},

						Instruction::THROW => {
							std::process::exit(self.pop() as i32);
						},

						Instruction::EXIT => {
							println!("\n[Program finished successfully]");
							std::process::exit(0);
						}
						// Exit actions
					}
				}
			}
			
		}
    }
    
	fn push(&mut self, value: u8) -> () {
		self.stack.push(value);
	}

	fn take(&mut self, index: u8) -> () {
		let tmp = self.stack.remove(index.into());
		self.push(tmp);
	}

	fn pop(&mut self) -> u8 {
		let a: Option<u8>=self.stack.pop();
		match a{
			None => {
				panic!("Failed to pop, ERROR 0x0003");
				0
			},
			Some(result) => {
				result
			}
		}
	}
}